# python core
import unittest
import time
import os

# from this project
from tqueue.base import get_db
from tqueue.client import queue_task, get_task
from tqueue.testutils import start_daemon, kill_daemon

PIDFILE = './pid.test'
LOGFILE = './test/test.log'


if os.environ.get('TQUEUE_DB_NAME') != 'tqueue_test':
    raise RuntimeError('Run tests with test.sh')


def cleandb():
    db = get_db()
    for coll in db.collection_names(include_system_collections=False):
        db[coll].delete_many({})
        db[coll].drop_indexes()


class TaskTest(unittest.TestCase):

    def setUp(self):
        start_daemon(PIDFILE, LOGFILE)
        cleandb()

    def tearDown(self):
        kill_daemon(PIDFILE)
        cleandb()

    def test_single_task(self):
        sessid, arg = 'mysess', 'myarg'
        taskid = queue_task(sessid, 'test.ns.tasks.echo_task', arg)
        task = get_task(taskid)
        time.sleep(0.5)
        task = get_task(taskid)
        self.assertEqual(task['status'], 'success')
        self.assertEqual(task['result'], [sessid, arg])

    def test_task_recovery(self):
        sessid, sleeptime = 'mysess', 2.0
        taskid = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime / 2.0)
        self.assertEqual(get_task(taskid)['status'], 'in_progress')
        kill_daemon(PIDFILE)
        self.assertEqual(get_task(taskid)['status'], 'queued')
        start_daemon(PIDFILE, LOGFILE)
        time.sleep(sleeptime / 2.0)
        self.assertEqual(get_task(taskid)['status'], 'in_progress')
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid)['status'], 'success')

    def test_two_same_session(self):
        sessid, sleeptime = 'mysess', 1
        taskid1 = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        taskid2 = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid1)['status'], 'success')
        self.assertEqual(get_task(taskid2)['status'], 'in_progress')
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid2)['status'], 'success')

    def test_single_session_pileup(self):
        sessid, sessid2, sleeptime = 'mysess', 'mysess2', 1
        taskid1 = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        taskid2 = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        taskid3 = queue_task(sessid2, 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid1)['status'], 'success')
        self.assertEqual(get_task(taskid3)['status'], 'success')
        self.assertEqual(get_task(taskid2)['status'], 'in_progress')
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid2)['status'], 'success')

    def test_session_rollover(self):
        sessid, sessid2, sessid3, sleeptime = 'mysess', 'mysess2', 'mysess3', 2
        # first, queue one task and sleep for a bit
        taskid1 = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime / 4.0)
        self.assertEqual(get_task(taskid1)['status'], 'in_progress')
        taskid2 = queue_task(sessid2, 'test.ns.tasks.sleepy_task', sleeptime)
        taskid3 = queue_task(sessid3, 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime / 2.0)
        # there are only 2 threads available and 3 sessions have tasks pending
        # at the same time. this means that only two of the 3 should be
        # running
        self.assertEqual(get_task(taskid1)['status'], 'in_progress')
        self.assertEqual(get_task(taskid2)['status'], 'in_progress')
        self.assertEqual(get_task(taskid3)['status'], 'queued')
        time.sleep((sleeptime / 2.0) + 0.5)
        # Now the first to that were allowd to run concurrently have
        # finished and the 3rd is now allowed to run
        self.assertEqual(get_task(taskid1)['status'], 'success')
        self.assertEqual(get_task(taskid2)['status'], 'success')
        self.assertEqual(get_task(taskid3)['status'], 'in_progress')
        time.sleep(sleeptime)
        # now we have enough time to finish the last one
        self.assertEqual(get_task(taskid3)['status'], 'success')

    def test_failed_task(self):
        sessid = 'mysess'
        taskid = queue_task(sessid, 'test.ns.tasks.error_task')
        time.sleep(1)
        self.assertEqual(get_task(taskid)['status'], 'failed')

    def test_failed_task_init(self):
        sessid, sessid2 = 'mysess', 'mysess2'
        taskid = queue_task(sessid, 'test.ns.fail.echo_task', 1)
        taskid2 = queue_task(sessid2, 'test.ns.fail.echo_task', 1)
        time.sleep(5)
        task1, task2 = get_task(taskid), get_task(taskid2)
        self.assertEqual(task1['status'], 'failed')
        self.assertEqual(task2['status'], 'failed')
        self.assertEqual(task1['error'][0], 'init() fail')
        self.assertEqual(task2['error'][0], 'init() fail')
        sleeptime = 1
        taskid3 = queue_task(
            sessid + 'asd', 'test.ns.tasks.sleepy_task', sleeptime)
        time.sleep(sleeptime + 0.5)
        self.assertEqual(get_task(taskid3)['status'], 'success')

    def test_set_payload_transform(self):
        sessid = 'mysess'
        taskid = queue_task(sessid, 'test.ns.custom_transform.echo_task', 1)
        time.sleep(0.5)
        self.assertEqual(get_task(taskid)['result'], {'hello': 'world'})

    def test_statuslog(self):
        sessid, sleeptime = 'mysess', 1
        taskid = queue_task(sessid, 'test.ns.tasks.sleepy_task', sleeptime)
        statuslist = lambda: [x[0] for x in get_task(taskid)['statuslog']]
        self.assertEqual(['queued'], statuslist())
        time.sleep(0.5)
        self.assertEqual(['queued', 'received', 'in_progress'], statuslist())
        time.sleep(0.7)
        self.assertEqual(
            ['queued', 'received', 'in_progress', 'success'], statuslist())
