# from python core
import os
import sys
import argparse
import signal
import multiprocessing
from multiprocessing import Semaphore, Queue, Process, Manager

from cputil import log

from .daemonutils import next_task, getpid, is_running, update_task
from .worker import worker
from .base import get_collection, get_db

log.setname('tqueue')


parser = argparse.ArgumentParser()
parser.add_argument(
    '-n', '--nprocs',
    help='Max number of concurrent sessions', type=int, required=True)
parser.add_argument(
    '-p', '--pidfile', help='Where to store the pid', type=str, required=False)
parser.add_argument(
    '-d', '--debug',
    help='If true print additional debug info', action='store_true',
    required=False)
parser.add_argument(
    '-f', '--forkserver',
    help='Use multiprocessing.set_start_server("forkserver"). Use on mac to '
    'prevent periodic python crash errors.',
    action='store_true', required=False)

# parser.add_argument(
#     '-s', '--single', help='Run the next task and exit', action='count')
# parser.add_argument(
#     '-t', '--task', help='ID of the task to run (will re-run if already '
#                          'completed or in error)', type=str)


args = parser.parse_args()

SEM = None

# maps the sessionid to the Process object
PROCESSES = {}

# thread safe list that will serve as an ordered list of available processes
# you can kill a process from this list without fear of stopping an
# already-running task
IDLE_SESSIONS = None

# if a session fails to initialize, it must add itself to this list
# and release the semaphore. then the daemon can clean it up.
FAILED_SESSIONS = None

# this maps a session id to a multiprocessing.Queue. this is how tasks
# are passed from the master process to an already running session process
TASK_QUEUES = {}


def session_exists(task):
    return task['session_id'] in TASK_QUEUES


def queue_task(task):
    update_task(task['_id'], 'received')
    TASK_QUEUES[task['session_id']].put(task)


def init_session_process(task):
    sessid = task['session_id']

    q = Queue()
    modulepath = '.'.join(task['name'].split('.')[:-1])
    p = Process(target=worker, args=(
        sessid, modulepath, q, IDLE_SESSIONS, FAILED_SESSIONS, SEM))
    TASK_QUEUES[sessid] = q
    PROCESSES[sessid] = p
    p.start()


def clean_failed_sessions():
    for sessid in FAILED_SESSIONS:
        log.warning('cleaning failed session {}'.format(sessid))
        kill_session(sessid)
        FAILED_SESSIONS.remove(sessid)


def kill_session(sessid):
    log.info('terminating session "{}"'.format(sessid))
    if sessid in PROCESSES:
        PROCESSES[sessid].terminate()
        del PROCESSES[sessid]
    else:
        log.warning('asking to kill a session not found in SESSIONS!!!')
    if sessid in TASK_QUEUES:
        del TASK_QUEUES[sessid]
    else:
        log.warning('asking to delete a task not found in TASK_QUEUES!!!')


def kill_session_if_needed():

    if len(PROCESSES) >= args.nprocs:
        # we know there must be at least one IDLE_SESSION because otherwise
        # the semaphore would not have been released
        kill_session(IDLE_SESSIONS.pop())


def acquire_sem():
    SEM.acquire()
    clean_failed_sessions()


def daemon():
    """
    will block in two places:
    1 - while waiting for a new task to show up
    2 - if nprocs tasks are concurrently in progress, until one becomes free
    """
    global IDLE_SESSIONS, FAILED_SESSIONS
    log.info('starting daemon with max of {} sessions.'.format(args.nprocs))
    log.info('will pick up tasks from: {}.{}'.format(
        get_db().name, get_collection(get_db()).name))
    with Manager() as manager:
        IDLE_SESSIONS, FAILED_SESSIONS = manager.list(), manager.list()
        while True:
            task = next_task()
            if args.debug:
                print('\nreceived {} for session {}'.format(
                    task['name'], task['session_id']))
                print('-----')
                print('IDLE_SESSIONS', IDLE_SESSIONS)
                print('FAILED_SESSIONS', FAILED_SESSIONS)
                print('PROCESSES', PROCESSES.keys())
                print('TASK_QUEUES', TASK_QUEUES.keys())
                print('-----')
            clean_failed_sessions()
            if not session_exists(task):
                acquire_sem()
                kill_session_if_needed()
                init_session_process(task)
            queue_task(task)


class AlreadyRunning(Exception):
    pass


def main():

    global SEM

    if args.forkserver:
        try:
            multiprocessing.set_start_method('forkserver')
        except AttributeError as e:
            # set_start_method is not present in older versions
            print('Not using multiprocessing start_method forkserver')
            print(e)

    if args.pidfile:
        if getpid(args.pidfile) and is_running(getpid(args.pidfile)):
            raise AlreadyRunning('daemon with pidfile {}'.format(args.pidfile))

        with open(args.pidfile, 'w') as fh:
            fh.write(str(os.getpid()))

    SEM = Semaphore(args.nprocs)
    pid = os.getpid()

    def graceful_exit(*_):
        # for some reason the child processes will also execute this function
        # so check that I am actually the parent before executing all of this
        if os.getpid() == pid:
            log.info('\nterminating {} processes'.format(len(PROCESSES)))
            for p in PROCESSES.values():
                p.terminate()
        sys.exit()

    signal.signal(signal.SIGINT, graceful_exit)
    signal.signal(signal.SIGTERM, graceful_exit)
    daemon()


if __name__ == '__main__':
    main()
